import logging
from laboro.logic.processor import Processor
from laboro.workflow.method import Method


class Action:
  """The ``laboro.workflow.action.Action`` object is a representation of a   ``laboro.workflow.step.Step`` action.

  Arguments:
    context: The ``laboro.context.Context`` instance used by the workflow.
    name: A string, specifying the action name.
    instance: A dictionary representation of ``instance`` as specified in the ``laboro.workflow.Workflow`` specification.
    when: Optional. A string representation of an expression thant can be evaluate as a boolean.
    loop: Optional. A string describing any iterable object.

  Returns:
    ``laboro.workflow.action.Action``
  """

  @property
  def iterable(self):
    """Get the *iterable* object on which loop, if any.
    This property is evaluated for each call from the ``loop`` attribute.
    """
    if self.loop is not None:
      return Processor().process(self.ctx, str(self.loop))
    return [self.loop]

  @property
  def runnable(self):
    """Get the condition in which the step is runnable.
    This property is evaluated for each call from the ``when`` attribute.
    """
    if self.when is not None:
      return Processor().process(self.ctx, str(self.when))
    return True

  def __init__(self, context, name, instance, when=None, loop=None):
    self.ctx = context
    self.name = name
    self.when = when
    self.loop = loop
    self.instance = instance

  def __enter__(self):
    return self

  def __exit__(self, kind, value, traceback):
    pass

  def _get_instance_args(self, instance):
    if "args" in instance:
      interpolated = instance["args"].copy()
      for key, value in interpolated.items():
        interpolated[key] = Processor().process_arg(self.ctx, str(value))
      return interpolated
    return dict()

  def run(self):
    """Run the action.
    If the ``runnable`` property evaluate to ``True``, calling the ``run()`` method will instantiate the instance when needed and run all ``methods`` within the instance methods list.
    The methods within instance methods list will be executed for each item in the ``iterable`` property.
    """
    for item in self.iterable:
      self.ctx.store.action_item = item
      if self.runnable:
        logging.getLogger().log_section("ACTION", self.name)
        cls = self.instance["class"]
        cls_args = self._get_instance_args(self.instance)
        if self.instance["instantiate"]:
          instance = self.ctx.instantiate(module=self.instance["module"],
                                          cls=cls,
                                          args=cls_args)
          self.ctx.put(self.instance["name"], instance)
        else:
          instance = self.ctx.get(self.instance["name"])
        for action_method in self.instance["methods"]:
          with Method(self.ctx, instance, **action_method) as method:
            method.run()
      else:
        msg = f"Skipping {self.name}: Condition not met: {self.when}"
        logging.getLogger().log_section("STEP", msg, level=logging.WARNING)



import sys
import logging
import argparse
from laboro.context import Context
from laboro.workflow import Workflow
from laboro.logger.manager import Manager as LogMgr
from laboro.vault import Vault


def run(workflow, context):
  try:
    with Workflow(name=workflow, context=context) as wkf:
      wkf.run()
  except Exception:
    pass


def main(workflows):
  context = Context(logger=LogMgr(Vault()))
  logging.getLogger().log_section("LABORO", "Bootstrapping")
  for workflow in workflows:
    logging.getLogger().vault.clear()
    run(workflow, context=context)


if __name__ == "__main__":
  parser = argparse.ArgumentParser(description="Run Laboro workflow",
                                   prog="laboro")
  parser.add_argument("-r", "--run",
                      metavar="workflow",
                      nargs="+",
                      required=True,
                      help="Run the specified workflows sequentially")
  args = parser.parse_args()
  if not args.run:
    parser.print_help()
    sys.exit(1)
  main(args.run)



import os
import logging
from laboro.error import LaboroError
from laboro.history import History
from laboro.workspace import Workspace
from laboro.context.store import Store
from laboro.config.manager import Manager as ConfigMgr
from laboro.module.manager import Manager as ModuleMgr
from laboro.error.handler import Handler as ErrorHandler


class Context:
  """The ``laboro.context.Context`` object manages all low level tasks for ``laboro.workflow.Workflow`` instances. It is instantiated once by the main **Laboro** process in a **Laboro** execution ans is passed from **Workflow** instances to **Workflow** instances.

  It loads the main **Laboro** configuration file and initiate all needed **Laboro** submodules.

  Arguments:
    logger: A ``laboro.logger.manager.Manager`` instance.

    main_config_file: Optional, a string specifying the main **Laboro** configuration file path. Defaults to ``/etc/laboro/laboro.yml``.

  Raises:
    ``laboro.error.LaboroError``: If the main **Laboro** configuration file is not found or invalid.

  Returns:
    ``laboro.context.Context``
  """
  def __init__(self, logger, main_config_file="/etc/laboro/laboro.yml"):
    self.delete_workspace_on_exit = True
    self.logger = logger
    self.store = Store()
    self.history = None
    self.modulemgr = ModuleMgr()
    self.configmgr = ConfigMgr(main_config_file)
    self.workspacedir = self.configmgr.get_parameter("main",
                                                     "$.laboro.workspacedir")
    self.histdir = self.configmgr.get_parameter("main",
                                                "$.laboro.histdir")
    self.workflowdir = self.configmgr.get_parameter("main",
                                                    "$.laboro.workflowdir")
    self.logdir = self.configmgr.get_parameter("main",
                                               "$.laboro.log.dir")
    self.loglevel = self.configmgr.get_parameter("main",
                                                 "$.laboro.log.level")
    self.workspace = None
    logging.getLogger().vault.clear()

  def reset(self, workflow_name, session):
    """Reset the ``laboro.context.Context`` instance with the new workflow configuration and session.

    Arguments:
      workflow_name: A string specifying the workflow name. The ``workflow_name`` will be used to retrieve the workflow configuration file in the``laboro.workflowdir`` specified in the main configuration file.
      session: A unique string defining a workflow session. Usually a string representation of a uuid.uuid4() instance.

    Raises:
      ``laboro.error.LaboroError``: For various reasons linked to the submodules instantiations.

    Returns:
      ``laboro.context.Context``: The ``Context`` instance reseted with the new parameters.
    """
    try:
      logging.getLogger().log_section("WORKFLOW",
                                      f"Started {workflow_name} / {session}")
      logging.getLogger().vault.clear()
      self.store.clear()
      self._configure_logger(workflow_name, session)
      self._set_workflow_config(workflow_name)
      self.delete_workspace_on_exit = self.configmgr.get_parameter("workflow", "$.workspace.delete_on_exit")
      self.workspace = Workspace(workspacedir=self.workspacedir,
                                 workflow=workflow_name,
                                 session=session)
      history_path = os.path.join(self.histdir, f"{workflow_name}.db")
      self.history = History(filename=history_path,
                             workflow=workflow_name,
                             session=session,
                             params=self.configmgr.workflow_config)
      self.history.enter()
      self.install_packages()
      return self
    except Exception as err:
      ErrorHandler().handle_error(workflow_name,
                                  session,
                                  err.__class__,
                                  str(err))

  def exit(self, name, session, kind, value):
    """Exit the ``Context`` instance.
    When the instance exit, it triggers the ``laboro.workspace.Workspace``, ``laboro.history.History`` instances exit and close the log linked to the ``laboro.workflow.Workflow`` instance associated with the ``Context`` instance. It also register the exit/error code.

    Arguments:
      name: A string specifying the ``Workflow`` name to close.
      session: A string specifying the ``Workflow`` session to close.
      kind: The kind of event that triggered the ``Context`` instance exit.
      value: The `value` of the event that triggered the ``Context`` instance exit.
    """
    if self.delete_workspace_on_exit:
      self.workspace.delete()
    self.history.exit(kind, value)
    ErrorHandler().handle_error(name, session, kind, value)
    self.logger.remove_file_handler(session)

  def _configure_logger(self, workflow_name, session):
    self.logger.add_file_handler(self.logdir, workflow_name, session)
    self.logger.set_log_level(self.loglevel)

  def _set_workflow_config(self, workflow_name):
    config_file_name = os.path.join(self.workflowdir, f"{workflow_name}.yml")
    logging.info(f"[+] Loading workflow_config: {config_file_name}")
    self.configmgr.workflow_config = config_file_name

  def install_packages(self):
    """Install all packages listed in the workflow configuration file.
    """
    packages = self.configmgr.get_parameter("workflow", "$.packages")
    for pkg in packages:
      self.modulemgr.install_package(pkg)

  def put(self, prop, value):
    """Store the ``prop`` variable and its value in the ``Context.store`` ``laboro.context.Store`` instance.

    Arguments:
      prop: A string representing the variable name.
      value: Any type, the value to store.

    Raises:
      ``TypeError``: When ``prop`` is not a string.
    """
    self.store.put(prop, value)

  def get(self, prop):
    """Retrieve the ``prop`` value from the ``Context.store`` ``laboro.context.Store`` instance.

    Arguments:
      prop: A string representing the variable name.

    Raises:
      ``TypeError``: When ``prop`` is not a string.
      ``AttributeError``: When ``prop`` does not exist within the Store instance.
    """
    return self.store.get(prop)

  def _register_class(self, module, cls):
    self.modulemgr.register_class_from_module(cls, module)

  def _register_instance_secrets(self, instance, args):
    if "args" in instance.specification and args is not None:
      class_args = instance.specification["args"]
      secret_keys = [arg["name"] for arg in class_args if arg["secret"]]
      secrets = [args[key] for key in args.keys() if key in secret_keys]
      list(map(logging.getLogger().vault.add, secrets))

  def _instantiate(self, module, cls, args):
    logging.info(f"[+] Object instantiation: {module}.{cls}")
    mod = self.modulemgr.get_class_from_module(cls=cls, module=module)
    instance = mod(context=self, args=args)
    self._register_instance_secrets(instance, args)
    return mod(context=self, args=args)

  def instantiate(self, module, cls, args):
    """Register the class to the ``laboro.module.ModuleMgr`` instance and return an instance of the class.

    Arguments:
      module: A string specifying the *Python* module name from each retrieve the class.
      cls: A string specifying the class name to instantiate.
      args: A `kwargs` that will be used to instantiate the class.

    Returns:
      An instance of the class ``module``.``cls`` instantiated with the ``args`` `kwargs` as init parameters.

    Raises:
      ``laboro.error.LaboroError``: When the specified module or class ar not available and when the specified ``args`` are incompatibles with the class constructor.
    """
    self._register_class(module, cls)
    return self._instantiate(module, cls, args)

  def register_method_secrets(self, instance, method, args):
    """Register the secrets from the specified arguments ``args``.
    Retrieve the name of the ``instance``'s ``method`` declared as *secret* and register their value in the ``laboro.Vault.vault`` embedded in the ``laboro.logger.LaboroLogger``.

    Arguments:
      instance: An instance of an object derived from the ``laboro.module.Module``.
      method: The ``instance`` method name from which retrieve the secret arguments list.
      args: A kwargs from which retrieve the secret value.

    Raises:
      ``laboro.error.LaboroError`` when the specified ``method`` does not exists.
    """
    if args is not None:
      try:
        method_spec = [meth for meth in instance.specification["methods"] if meth["name"] == method][0]
        if "args" in method_spec:
          method_args = method_spec["args"]
          secret_keys = [arg["name"] for arg in method_args if arg["secret"]]
          secrets = [args[key] for key in args.keys() if key in secret_keys]
          list(map(logging.getLogger().vault.add, secrets))
      except IndexError as err:
        raise LaboroError(f"UnknownMethodError: Unknown method {instance.__class__.__name__}.{method}") from err



import re
from laboro.logic.parser import Parser


class Processor:
  @staticmethod
  def process(context, statement):
    word_ptn = re.compile(r"\$store\$\w+")
    store_ptn = re.compile(r"^\$store\$")
    vars = re.findall(word_ptn, statement)
    values = [str(context.get(re.sub(store_ptn, "", var))) for var in vars]
    for var, value in zip(vars, values):
      var_ptn = re.compile(var.replace("$", r"\$"))
      statement = re.sub(var_ptn, value, statement)
    if context.store.step_item is not None:
      pattern = re.compile(r"\$step_item\$")
      statement = re.sub(pattern, str(context.store.step_item), statement)
    if context.store.action_item is not None:
      pattern = re.compile(r"\$action_item\$")
      statement = re.sub(pattern, str(context.store.action_item), statement)
    if context.store.method_item is not None:
      pattern = re.compile(r"\$method_item\$")
      statement = re.sub(pattern, str(context.store.method_item), statement)
    return Parser().eval(statement)

  @staticmethod
  def process_arg(context, value):
    store_ptn = re.compile(r"\$store\$\w+")
    for key in re.findall(store_ptn, value):
      val = context.get(key.replace("$store$", ""))
      value = value.replace(key, str(val))
    if context.store.step_item is not None:
      pattern = re.compile(r"\$step_item\$")
      value = re.sub(pattern, str(context.store.step_item), value)
    if context.store.action_item is not None:
      pattern = re.compile(r"\$action_item\$")
      value = re.sub(pattern, str(context.store.action_item), value)
    if context.store.method_item is not None:
      pattern = re.compile(r"\$method_item\$")
      value = re.sub(pattern, str(context.store.method_item), value)
    return Parser().literal_eval(value)



import logging


class LaboroLogger(logging.Logger):
  """The ``laboro.logger.LaboroLogger`` overrides the ``logging.Logger`` class.

  It adds a ``laboro.vault.Vault`` instance to the logger object allowing secret management.

  Any secret stored in the vault instance will be redacted before logging.
  """
  levels = {logging.DEBUG: logging.debug,
            logging.WARNING: logging.warning,
            logging.INFO: logging.info,
            logging.ERROR: logging.error,
            logging.CRITICAL: logging.critical}

  @staticmethod
  def log_line(level=logging.INFO):
    LaboroLogger.levels[level]("-" * 80)

  @staticmethod
  def log_section(prefix, message, level=logging.INFO):
    LaboroLogger.log_line(level=level)
    LaboroLogger.levels[level](f"[{prefix}] {message}")
    # LaboroLogger.log_line(level=level)

  def __init__(self, name, level=logging.NOTSET):
    self.vault = None
    super(LaboroLogger, self).__init__(name, level)

  def set_vault(self, vault):
    """Set the ``laboro.vault.Vault`` instance used to prevent secrets to be logged."""
    self.vault = vault

  def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
    """Override logging.Logger.makeRecord.
    """
    msg = self.vault.protect(msg)
    return super(LaboroLogger, self).makeRecord(name, level, fn, lno, msg, args, exc_info, func, extra, sinfo)



import logging
from ast import literal_eval
from asteval import Interpreter


class Parser:
  @staticmethod
  def eval(expression):
    try:
      return Interpreter().eval(expression,
                                show_errors=False,
                                raise_errors=True)
    except SyntaxError as err:
      logging.critical(f"Invalid syntax for expression: {expression}")
      raise err

  @staticmethod
  def literal_eval(expression):
    try:
      return literal_eval(expression)
    except (ValueError, SyntaxError):
      return expression



import logging
import traceback


class Handler:
  """The ``laboro.error.handler.Handler`` is a singleton that logs every event given to its ``handle_error()`` method.
  """

  @staticmethod
  def handle_error(name, session, kind, value):
    """Log the given error with its traceback and exit.

    Arguments:
      name: A string specifying the ``laboro.workflow.Workflow`` instance name.
      session: A string specifying the ``laboro.workflow.Workflow`` instance session.
      kind: The `kind` of event. The type of `kind` is variable.
      value: The value of the event. The type of `value` depends on `kind`.
    """
    level = logging.INFO
    exit_msg = "Exited"
    exit_code = 0
    if kind == SystemExit:
      if value.code is not None:
        exit_code = value.code
    elif kind is not None:
      level = logging.CRITICAL
      exit_code = f"{kind.__name__}: {value}"
    exit_msg += f" {name} / {session} with {exit_code}"
    logging.getLogger().log_section("WORKFLOW", exit_msg, level=level)
    if kind is not None and kind != SystemExit:
      if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
        traceback.print_exc()



import os
import time
import logging
import sqlite3
from datetime import datetime
from prettytable import PrettyTable


class History:
  """The ``laboro.history.History`` class provides all methods necessary to manage execution history. The history file is a `sqlite` database.

  Arguments:
      filename: A string specifying the full path to the history file to use
      workflow: A string defining the workflow name.
      session: A unique string defining the workflow execution session.
      params: A dict object listing all workflow parameters and their values.
      retention: An integer defining how many executions an history record should keep in the history file.

  Returns:
    ``laboro.history.History``: An history object.

  Raises:
    ``sqlite3.Error``: Any sqlite3 error may be raised when such an error occurs.
  """

  def __init__(self, filename, workflow, session, params, retention=90):
    self.hist_filename = filename
    self.workflow = workflow
    self.params = str(params)
    self.start = time.time()
    self.session = session
    self.retention = retention

  def _create_db(self):
    logging.info(f"[+] Checking history database: {self.hist_filename}")
    if not os.path.exists(self.hist_filename):
      with sqlite3.connect(self.hist_filename) as conn:
        cursor = conn.cursor()
        cursor.execute("CREATE TABLE executions (workflow, session TEXT, start REAL, end REAL, duration TEXT, params TEXT, exit_code INTEGER)")
        conn.commit()

  def _end_exec(self, exit_code):
    logging.info(f"[+] Registering end of session: [{self.session}]: {exit_code}")
    end = time.time()
    delta = datetime.fromtimestamp(end) - datetime.fromtimestamp(self.start)
    result = [end,
              self._format_time_delta(delta),
              exit_code,
              self.session]
    with sqlite3.connect(self.hist_filename) as conn:
      cursor = conn.cursor()
      cursor.execute("update executions set end=?, duration=?, exit_code=? where session=?", result)
      conn.commit()

  def _purge_history(self):
    """Purge history database."""
    logging.info(f"[+] Purging history: Keeping last {self.retention} executions")
    with sqlite3.connect(self.hist_filename) as conn:
      cursor = conn.cursor()
      cursor.execute("delete from executions where session not in (select session from executions order by end DESC limit ?)", [self.retention])
      conn.commit()

  def _format_time_delta(self, tdelta):
    """Take a timedelta object and formats it for humans.
    From https://gist.github.com/dhrrgn/7255361
    """
    delta = dict(days=tdelta.days)
    delta["hrs"], rem = divmod(tdelta.seconds, 3600)
    delta["min"], delta["sec"] = divmod(rem, 60)
    if delta["min"] == 0:
      fmt = "{sec} sec"
    elif delta["hrs"] == 0:
      fmt = "{min} min {sec} sec"
    elif delta["days"] == 0:
      fmt = "{hrs} hr(s) {min} min {sec} sec"
    else:
      fmt = "{days} day(s) {hrs} hr(s) {min} min {sec} sec"
    return fmt.format(**delta)

  def enter(self):
    """Start the ``laboro.history.History`` instance.
    Create the history file database if needed, purge the associated ``laboro.workflow.Workflow`` history execution and record the present execution parameters.

    Registered parameters are:
    - The workflow name
    - The workflow session
    - The start date timestamp
    - The workflow init parameters
    """
    self._create_db()
    self._purge_history()
    record = [self.workflow,
              self.session,
              self.start,
              self.params]
    with sqlite3.connect(self.hist_filename) as conn:
      cursor = conn.cursor()
      cursor.execute("insert into executions (workflow, session, start, params) values (?, ?, ?, ?)", record)
      conn.commit()

  def exit(self, kind, value):
    """Exit the ``laboro.history.History`` instance.
    When exiting, the ``laboro.history.History`` instance records the end date timestamp, the time delta between execution start and end the exit code type and its value.

    Arguments:
      kind: The kind of event that triggered the exit. Its type is variable.
      value: The value of the event that triggered the exit. Its type depends on the ``kind`` type.
    """
    exit_code = 0
    if kind == SystemExit:
      if value.code is not None:
        exit_code = value.code
    elif kind is not None:
      exit_code = f"{kind.__name__}: {value}"
    self._end_exec(exit_code)

  def read_history(self, workflow, num_exec):
    """Displays last ``num_exec`` history lines for specified ``workflow``.

    At this point, this method is not used by **Laboro** itself. It may be accessible in a future version of **Laboro**.

    Arguments:
      workflow: A string representing the workflow name to search for.
      num_exec: An integer representing the number of executions to retrieve.

    Returns:
      ``str``: A string representation of the history displayable as a PrettyTable.
    """
    table = PrettyTable()
    table.field_names = ["Workflow",
                         "Session",
                         "Start",
                         "End",
                         "Duration",
                         "Params",
                         "Exit"]
    with sqlite3.connect(self.hist_filename) as conn:
      cursor = conn.cursor()
      cursor.execute("select * from executions where workflow=? order by start desc limit ?", [workflow, num_exec])
      table.add_rows(cursor.fetchall())
      return table.get_string()



import uuid
import logging
from laboro.workflow.step import Step


class Workflow:
  """The ``laboro.workflow.Workflow`` object is the main class for the workflow representation.
  It load all configuration needed, set alk objects such History, Vault and workspace and run according to its configuration.

  The Workflow object provides a runtime context that will handle log, history, vault, and workspace, etc.

  Arguments:
    name: A string representing the workflow name.
    context: A ``laboro.context.Context`` instance.

  Returns:
    ``laboro.workflow.Workflow``: A Workflow object.

  ..  code-block:: python

      from laboro.vault import Vault()
      from laboro.log.manager import Manager as LogMgr
      from laboro.context import Context
      from laboro.workflow import Workflow

      logmgr = LogMgr(vault=Vault())
      context = Context(logger=logmgr)
      with Workflow(name="my_workflow", context=context) as wkf:
        wkf.run()
        ...
  """
  def __init__(self, name, context):
    self.name = name
    self.session = str(uuid.uuid4())
    self.ctx = context.reset(self.name, self.session)

  def __enter__(self):
    return self

  def __exit__(self, kind, value, traceback):
    self.ctx.exit(self.name, self.session, kind, value)

  def run(self):
    """Run the workflow."""
    for wkf_step in self.ctx.configmgr.get_parameter("workflow", "$.steps"):
      with Step(self.ctx, **wkf_step) as step:
        step.run()
    logging.getLogger().log_line()



import os
import logging
from jsonpath import JSONPath
from laboro.error import LaboroError
from laboro.validator import Validator


class Manager:
  """The ``laboro.config.manager.Manager`` object loads and check all configuration files needed by  **Laboro** to operate.

  Arguments:
    `main_config`: A string representing the main **Laboro** configuration filename.

  Returns:
    ``laboro.config.manager.Manager``: The **Laboro** configuration manager.

  Raises:
    ``laboro.error.LaboroError``: When one of the configuration files are not found.
  """

  @property
  def main_config(self):
    """Get the ``main_config`` as a dict."""
    return self._main_config

  @property
  def workflow_config(self):
    """Get the ``workflow_config`` as a dict."""
    return self._workflow_config

  @main_config.setter
  def main_config(self, filename):
    """Set the main_config from a YAML file.

    Arguments:
      filename: A string specifying the YAML file to load the configuration from."""
    schema = os.path.join(os.path.dirname(__file__),
                          "schema",
                          "main.yml")
    self._main_config = Validator().validate_from_files(schema=schema,
                                                        instance=filename)

  @workflow_config.setter
  def workflow_config(self, filename):
    """Set the workflow_config from a YAML file.

    Arguments:
      filename: A string specifying the YAML file to load the configuration from."""
    schema = os.path.join(os.path.dirname(__file__),
                          "schema",
                          "workflow.yml")
    self._workflow_config = Validator().validate_from_files(schema=schema,
                                                            instance=filename)

  def __init__(self, main_config):
    self.main_config = main_config
    self._workflow_config = dict()

  def get_parameter(self, level, param):
    """Returns the configuration parameter specified by ``param`` from the configuration ``level``.

    Arguments:
      level: A string that specify the level from which retrieve the parameter. Must be one of ``main`` or ``workflow``.
      param: A string that specify the parameter name as described by the JSONPath spec. The ``param`` argument is in `dotted` format.

    Returns:
      Any type, depending on `YAML` file from which the configuration is generated.

    Raises:
      ``laboro.error.LaboroError``: When the specified parameter is unknown or ``level`` value is not valid.

    ..  code-block:: python

        manager = Manager(main_conf)
        manager.get_parameter("workflow",
                              "$.steps[0].actions[0].instance.args.my_arg")
    """
    levels = {"main": self.main_config,
              "workflow": self.workflow_config}
    try:
      config = levels[level]
      try:
        return JSONPath(param).parse(config)[0]
      except IndexError as err:
        msg = f"UnknownParameterError: [{level}]: {param}"
        logging.critical(msg)
        raise LaboroError(msg) from err
    except KeyError as err:
      msg = f"BadConfLevelError: Bad configuration level: {level}"
      logging.critical(msg)
      raise LaboroError(msg) from err
